@@ -4,13 +4,23 @@
 {-# LANGUAGE DeriveDataTypeable         #-}
 {-# LANGUAGE FlexibleContexts           #-}
 {-# LANGUAGE FlexibleInstances          #-}
-{-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE MultiParamTypeClasses      #-}
 {-# LANGUAGE OverloadedStrings          #-}
 {-# LANGUAGE PolyKinds                  #-}
+{-# LANGUAGE RankNTypes                 #-}
 {-# LANGUAGE ScopedTypeVariables        #-}
+{-# LANGUAGE TupleSections              #-}
 {-# LANGUAGE TypeFamilies               #-}
 {-# LANGUAGE TypeOperators              #-}
+{-# LANGUAGE UndecidableInstances       #-}
+
+#if MIN_VERSION_base(4,9,0) && __GLASGOW_HASKELL__ >= 802
+#define HAS_TYPE_ERROR
+#endif
+
+#ifdef HAS_TYPE_ERROR
+{-# LANGUAGE UndecidableInstances       #-}
+#endif
 
 #include "overlapping-compat.h"
 
@@ -24,45 +34,58 @@ module Servant.Server.Internal
   , module Servant.Server.Internal.ServantErr
   ) where
 
+import           Control.Monad              (join, when)
 import           Control.Monad.Trans        (liftIO)
 import           Control.Monad.Trans.Resource (runResourceT)
 import qualified Data.ByteString            as B
+import qualified Data.ByteString.Builder    as BB
 import qualified Data.ByteString.Char8      as BC8
 import qualified Data.ByteString.Lazy       as BL
-import           Data.Maybe                 (fromMaybe, mapMaybe)
+import           Data.Maybe                 (fromMaybe, mapMaybe,
+                                             isNothing, maybeToList)
 import           Data.Either                (partitionEithers)
-import           Data.String                (fromString)
+import           Data.String                (IsString (..))
 import           Data.String.Conversions    (cs, (<>))
-import           Data.Tagged                (Tagged(..), untag)
+import           Data.Tagged                (Tagged(..), retag, untag)
 import qualified Data.Text                  as T
 import           Data.Typeable
 import           GHC.TypeLits               (KnownNat, KnownSymbol, natVal,
                                              symbolVal)
 import           Network.HTTP.Types         hiding (Header, ResponseHeaders)
+import qualified Network.HTTP.Media         as NHM
 import           Network.Socket             (SockAddr)
-import           Network.Wai                (Application, Request, Response,
+import           Network.Wai                (Application, Request,
                                              httpVersion, isSecure,
                                              lazyRequestBody,
                                              rawQueryString, remoteHost,
                                              requestHeaders, requestMethod,
-                                             responseLBS, vault)
+                                             responseLBS, responseStream,
+                                             vault)
 import           Prelude                    ()
 import           Prelude.Compat
 import           Web.HttpApiData            (FromHttpApiData, parseHeader,
                                              parseQueryParam,
                                              parseUrlPieceMaybe,
                                              parseUrlPieces)
-import           Servant.API                 ((:<|>) (..), (:>), BasicAuth, Capture,
+import           Servant.API                 ((:<|>) (..), (:>), BasicAuth, Capture',
                                               CaptureAll, Verb, EmptyAPI,
                                               ReflectMethod(reflectMethod),
-                                              IsSecure(..), Header, QueryFlag,
-                                              QueryParam, QueryParams, Raw,
-                                              RemoteHost, ReqBody, Vault,
-                                              WithNamedContext)
+                                              IsSecure(..), Header', QueryFlag,
+                                              QueryParam', QueryParams, Raw,
+                                              RemoteHost, ReqBody', Vault,
+                                              WithNamedContext,
+                                              Description, Summary,
+                                              Accept(..),
+                                              FramingRender(..), Stream,
+                                              StreamGenerator(..), ToStreamGenerator(..),
+                                              BoundaryStrategy(..),
+                                              If, SBool (..), SBoolI (..))
+import           Servant.API.Modifiers       (unfoldRequestArgument, RequestArgument, FoldRequired, FoldLenient)
 import           Servant.API.ContentTypes    (AcceptHeader (..),
                                               AllCTRender (..),
                                               AllCTUnrender (..),
                                               AllMime,
+                                              MimeRender(..),
                                               canHandleAcceptH)
 import           Servant.API.ResponseHeaders (GetHeaders, Headers, getHeaders,
                                               getResponse)
@@ -74,6 +97,9 @@ import           Servant.Server.Internal.Router
 import           Servant.Server.Internal.RoutingApplication
 import           Servant.Server.Internal.ServantErr
 
+#ifdef HAS_TYPE_ERROR
+import GHC.TypeLits (TypeError, ErrorMessage (..))
+#endif
 
 class HasServer api context where
   type ServerT api (m :: * -> *) :: *
@@ -84,6 +110,13 @@ class HasServer api context where
     -> Delayed env (Server api)
     -> Router env
 
+  hoistServerWithContext
+      :: Proxy api
+      -> Proxy context
+      -> (forall x. m x -> n x)
+      -> ServerT api m
+      -> ServerT api n
+
 type Server api = ServerT api Handler
 
 -- * Instances
@@ -108,6 +141,11 @@ instance (HasServer a context, HasServer b context) => HasServer (a :<|> b) cont
     where pa = Proxy :: Proxy a
           pb = Proxy :: Proxy b
 
+  -- | This is better than 'enter', as it's tailor made for 'HasServer'.
+  hoistServerWithContext _ pc nt (a :<|> b) =
+    hoistServerWithContext (Proxy :: Proxy a) pc nt a :<|>
+    hoistServerWithContext (Proxy :: Proxy b) pc nt b
+
 -- | If you use 'Capture' in one of the endpoints for your API,
 -- this automatically requires your server-side handler to be a function
 -- that takes an argument of the type specified by the 'Capture'.
@@ -126,11 +164,13 @@ instance (HasServer a context, HasServer b context) => HasServer (a :<|> b) cont
 -- >   where getBook :: Text -> Handler Book
 -- >         getBook isbn = ...
 instance (KnownSymbol capture, FromHttpApiData a, HasServer api context)
-      => HasServer (Capture capture a :> api) context where
+      => HasServer (Capture' mods capture a :> api) context where
 
-  type ServerT (Capture capture a :> api) m =
+  type ServerT (Capture' mods capture a :> api) m =
      a -> ServerT api m
 
+  hoistServerWithContext _ pc nt s = hoistServerWithContext (Proxy :: Proxy api) pc nt . s
+
   route Proxy context d =
     CaptureRouter $
         route (Proxy :: Proxy api)
@@ -157,15 +197,17 @@ instance (KnownSymbol capture, FromHttpApiData a, HasServer api context)
 -- > server = getSourceFile
 -- >   where getSourceFile :: [Text] -> Handler Book
 -- >         getSourceFile pathSegments = ...
-instance (KnownSymbol capture, FromHttpApiData a, HasServer sublayout context)
-      => HasServer (CaptureAll capture a :> sublayout) context where
+instance (KnownSymbol capture, FromHttpApiData a, HasServer api context)
+      => HasServer (CaptureAll capture a :> api) context where
 
-  type ServerT (CaptureAll capture a :> sublayout) m =
-    [a] -> ServerT sublayout m
+  type ServerT (CaptureAll capture a :> api) m =
+    [a] -> ServerT api m
+
+  hoistServerWithContext _ pc nt s = hoistServerWithContext (Proxy :: Proxy api) pc nt . s
 
   route Proxy context d =
     CaptureAllRouter $
-        route (Proxy :: Proxy sublayout)
+        route (Proxy :: Proxy api)
               context
               (addCapture d $ \ txts -> case parseUrlPieces txts of
                  Left _  -> delayedFail err400
@@ -179,16 +221,6 @@ allowedMethodHead method request = method == methodGet && requestMethod request
 allowedMethod :: Method -> Request -> Bool
 allowedMethod method request = allowedMethodHead method request || requestMethod request == method
 
-processMethodRouter :: Maybe (BL.ByteString, BL.ByteString) -> Status -> Method
-                    -> Maybe [(HeaderName, B.ByteString)]
-                    -> Request -> RouteResult Response
-processMethodRouter handleA status method headers request = case handleA of
-  Nothing -> FailFatal err406 -- this should not happen (checked before), so we make it fatal if it does
-  Just (contentT, body) -> Route $ responseLBS status hdrs bdy
-    where
-      bdy = if allowedMethodHead method request then "" else body
-      hdrs = (hContentType, cs contentT) : (fromMaybe [] headers)
-
 methodCheck :: Method -> Request -> DelayedIO ()
 methodCheck method request
   | allowedMethod method request = return ()
@@ -207,41 +239,32 @@ acceptCheck proxy accH
   | otherwise                                  = delayedFail err406
 
 methodRouter :: (AllCTRender ctypes a)
-             => Method -> Proxy ctypes -> Status
-             -> Delayed env (Handler a)
+             => (b -> ([(HeaderName, B.ByteString)], a))
+             -> Method -> Proxy ctypes -> Status
+             -> Delayed env (Handler b)
              -> Router env
-methodRouter method proxy status action = leafRouter route'
+methodRouter splitHeaders method proxy status action = leafRouter route'
   where
     route' env request respond =
           let accH = fromMaybe ct_wildcard $ lookup hAccept $ requestHeaders request
           in runAction (action `addMethodCheck` methodCheck method request
                                `addAcceptCheck` acceptCheck proxy accH
                        ) env request respond $ \ output -> do
-               let handleA = handleAcceptH proxy (AcceptHeader accH) output
-               processMethodRouter handleA status method Nothing request
-
-methodRouterHeaders :: (GetHeaders (Headers h v), AllCTRender ctypes v)
-                    => Method -> Proxy ctypes -> Status
-                    -> Delayed env (Handler (Headers h v))
-                    -> Router env
-methodRouterHeaders method proxy status action = leafRouter route'
-  where
-    route' env request respond =
-          let accH    = fromMaybe ct_wildcard $ lookup hAccept $ requestHeaders request
-          in runAction (action `addMethodCheck` methodCheck method request
-                               `addAcceptCheck` acceptCheck proxy accH
-                       ) env request respond $ \ output -> do
-                let headers = getHeaders output
-                    handleA = handleAcceptH proxy (AcceptHeader accH) (getResponse output)
-                processMethodRouter handleA status method (Just headers) request
+               let (headers, b) = splitHeaders output
+               case handleAcceptH proxy (AcceptHeader accH) b of
+                 Nothing -> FailFatal err406 -- this should not happen (checked before), so we make it fatal if it does
+                 Just (contentT, body) ->
+                      let bdy = if allowedMethodHead method request then "" else body
+                      in Route $ responseLBS status ((hContentType, cs contentT) : headers) bdy
 
 instance OVERLAPPABLE_
          ( AllCTRender ctypes a, ReflectMethod method, KnownNat status
          ) => HasServer (Verb method status ctypes a) context where
 
   type ServerT (Verb method status ctypes a) m = m a
+  hoistServerWithContext _ _ nt s = nt s
 
-  route Proxy _ = methodRouter method (Proxy :: Proxy ctypes) status
+  route Proxy _ = methodRouter ([],) method (Proxy :: Proxy ctypes) status
     where method = reflectMethod (Proxy :: Proxy method)
           status = toEnum . fromInteger $ natVal (Proxy :: Proxy status)
 
@@ -251,11 +274,76 @@ instance OVERLAPPING_
          ) => HasServer (Verb method status ctypes (Headers h a)) context where
 
   type ServerT (Verb method status ctypes (Headers h a)) m = m (Headers h a)
+  hoistServerWithContext _ _ nt s = nt s
 
-  route Proxy _ = methodRouterHeaders method (Proxy :: Proxy ctypes) status
+  route Proxy _ = methodRouter (\x -> (getHeaders x, getResponse x)) method (Proxy :: Proxy ctypes) status
     where method = reflectMethod (Proxy :: Proxy method)
           status = toEnum . fromInteger $ natVal (Proxy :: Proxy status)
 
+
+instance OVERLAPPABLE_
+         ( MimeRender ctype a, ReflectMethod method,
+           FramingRender framing ctype, ToStreamGenerator f a
+         ) => HasServer (Stream method framing ctype (f a)) context where
+
+  type ServerT (Stream method framing ctype (f a)) m = m (f a)
+  hoistServerWithContext _ _ nt s = nt s
+
+  route Proxy _ = streamRouter ([],) method (Proxy :: Proxy framing) (Proxy :: Proxy ctype)
+      where method = reflectMethod (Proxy :: Proxy method)
+
+instance OVERLAPPING_
+         ( MimeRender ctype a, ReflectMethod method,
+           FramingRender framing ctype, ToStreamGenerator f a,
+           GetHeaders (Headers h (f a))
+         ) => HasServer (Stream method framing ctype (Headers h (f a))) context where
+
+  type ServerT (Stream method framing ctype (Headers h (f a))) m = m (Headers h (f a))
+  hoistServerWithContext _ _ nt s = nt s
+
+  route Proxy _ = streamRouter (\x -> (getHeaders x, getResponse x)) method (Proxy :: Proxy framing) (Proxy :: Proxy ctype)
+      where method = reflectMethod (Proxy :: Proxy method)
+
+
+streamRouter :: (MimeRender ctype a, FramingRender framing ctype, ToStreamGenerator f a) =>
+                (b -> ([(HeaderName, B.ByteString)], f a))
+             -> Method
+             -> Proxy framing
+             -> Proxy ctype
+             -> Delayed env (Handler b)
+             -> Router env
+streamRouter splitHeaders method framingproxy ctypeproxy action = leafRouter $ \env request respond ->
+          let accH    = fromMaybe ct_wildcard $ lookup hAccept $ requestHeaders request
+              cmediatype = NHM.matchAccept [contentType ctypeproxy] accH
+              accCheck = when (isNothing cmediatype) $ delayedFail err406
+              contentHeader = (hContentType, NHM.renderHeader . maybeToList $ cmediatype)
+          in runAction (action `addMethodCheck` methodCheck method request
+                               `addAcceptCheck` accCheck
+                       ) env request respond $ \ output ->
+                let (headers, fa) = splitHeaders output
+                    k = getStreamGenerator . toStreamGenerator $ fa in
+                Route $ responseStream status200 (contentHeader : headers) $ \write flush -> do
+                      write . BB.lazyByteString $ header framingproxy ctypeproxy
+                      case boundary framingproxy ctypeproxy of
+                           BoundaryStrategyBracket f ->
+                                    let go x = let bs = mimeRender ctypeproxy x
+                                                   (before, after) = f bs
+                                               in write (   BB.lazyByteString before
+                                                         <> BB.lazyByteString bs
+                                                         <> BB.lazyByteString after) >> flush
+                                    in k go go
+                           BoundaryStrategyIntersperse sep -> k
+                             (\x -> do
+                                write . BB.lazyByteString . mimeRender ctypeproxy $ x
+                                flush)
+                             (\x -> do
+                                write . (BB.lazyByteString sep <>) . BB.lazyByteString . mimeRender ctypeproxy $ x
+                                flush)
+                           BoundaryStrategyGeneral f ->
+                                    let go = (>> flush) . write . BB.lazyByteString . f . mimeRender ctypeproxy
+                                    in k go go
+                      write . BB.lazyByteString $ trailer framingproxy ctypeproxy
+
 -- | If you use 'Header' in one of the endpoints for your API,
 -- this automatically requires your server-side handler to be a function
 -- that takes an argument of the type specified by 'Header'.
@@ -276,27 +364,39 @@ instance OVERLAPPING_
 -- > server = viewReferer
 -- >   where viewReferer :: Referer -> Handler referer
 -- >         viewReferer referer = return referer
-instance (KnownSymbol sym, FromHttpApiData a, HasServer api context)
-      => HasServer (Header sym a :> api) context where
+instance
+  (KnownSymbol sym, FromHttpApiData a, HasServer api context
+  , SBoolI (FoldRequired mods), SBoolI (FoldLenient mods)
+  )
+  => HasServer (Header' mods sym a :> api) context where
+------
+  type ServerT (Header' mods sym a :> api) m =
+    RequestArgument mods a -> ServerT api m
 
-  type ServerT (Header sym a :> api) m =
-    Maybe a -> ServerT api m
+  hoistServerWithContext _ pc nt s = hoistServerWithContext (Proxy :: Proxy api) pc nt . s
 
   route Proxy context subserver = route (Proxy :: Proxy api) context $
       subserver `addHeaderCheck` withRequest headerCheck
     where
-      headerName = symbolVal (Proxy :: Proxy sym)
+      headerName :: IsString n => n
+      headerName = fromString $ symbolVal (Proxy :: Proxy sym)
+
+      headerCheck :: Request -> DelayedIO (RequestArgument mods a)
       headerCheck req =
-        case lookup (fromString headerName) (requestHeaders req) of
-          Nothing -> return Nothing
-          Just txt ->
-            case parseHeader txt of
-              Left e -> delayedFailFatal err400
-                  { errBody = cs $ "Error parsing header "
-                                   <> fromString headerName
-                                   <> " failed: " <> e
-                  }
-              Right header -> return $ Just header
+          unfoldRequestArgument (Proxy :: Proxy mods) errReq errSt mev
+        where
+          mev :: Maybe (Either T.Text a)
+          mev = fmap parseHeader $ lookup headerName (requestHeaders req)
+
+          errReq = delayedFailFatal err400
+            { errBody = "Header " <> headerName <> " is required"
+            }
+
+          errSt e = delayedFailFatal err400
+              { errBody = cs $ "Error parsing header "
+                               <> headerName
+                               <> " failed: " <> e
+              }
 
 -- | If you use @'QueryParam' "author" Text@ in one of the endpoints for your API,
 -- this automatically requires your server-side handler to be a function
@@ -319,31 +419,41 @@ instance (KnownSymbol sym, FromHttpApiData a, HasServer api context)
 -- >   where getBooksBy :: Maybe Text -> Handler [Book]
 -- >         getBooksBy Nothing       = ...return all books...
 -- >         getBooksBy (Just author) = ...return books by the given author...
-instance (KnownSymbol sym, FromHttpApiData a, HasServer api context)
-      => HasServer (QueryParam sym a :> api) context where
+instance
+  ( KnownSymbol sym, FromHttpApiData a, HasServer api context
+  , SBoolI (FoldRequired mods), SBoolI (FoldLenient mods)
+  )
+  => HasServer (QueryParam' mods sym a :> api) context where
+------
+  type ServerT (QueryParam' mods sym a :> api) m =
+    RequestArgument mods a -> ServerT api m
 
-  type ServerT (QueryParam sym a :> api) m =
-    Maybe a -> ServerT api m
+  hoistServerWithContext _ pc nt s = hoistServerWithContext (Proxy :: Proxy api) pc nt . s
 
   route Proxy context subserver =
     let querytext req = parseQueryText $ rawQueryString req
+        paramname = cs $ symbolVal (Proxy :: Proxy sym)
+
+        parseParam :: Request -> DelayedIO (RequestArgument mods a)
         parseParam req =
-          case lookup paramname (querytext req) of
-            Nothing       -> return Nothing -- param absent from the query string
-            Just Nothing  -> return Nothing -- param present with no value -> Nothing
-            Just (Just v) ->
-              case parseQueryParam v of
-                  Left e -> delayedFailFatal err400
-                      { errBody = cs $ "Error parsing query parameter "
-                                       <> paramname <> " failed: " <> e
-                      }
+            unfoldRequestArgument (Proxy :: Proxy mods) errReq errSt mev
+          where
+            mev :: Maybe (Either T.Text a)
+            mev = fmap parseQueryParam $ join $ lookup paramname $ querytext req
+
+            errReq = delayedFailFatal err400
+              { errBody = cs $ "Query parameter " <> paramname <> " is required"
+              }
+
+            errSt e = delayedFailFatal err400
+              { errBody = cs $ "Error parsing query parameter "
+                               <> paramname <> " failed: " <> e
+              }
 
-                  Right param -> return $ Just param
         delayed = addParameterCheck subserver . withRequest $ \req ->
                     parseParam req
 
     in route (Proxy :: Proxy api) context delayed
-    where paramname = cs $ symbolVal (Proxy :: Proxy sym)
 
 -- | If you use @'QueryParams' "authors" Text@ in one of the endpoints for your API,
 -- this automatically requires your server-side handler to be a function
@@ -370,6 +480,8 @@ instance (KnownSymbol sym, FromHttpApiData a, HasServer api context)
   type ServerT (QueryParams sym a :> api) m =
     [a] -> ServerT api m
 
+  hoistServerWithContext _ pc nt s = hoistServerWithContext (Proxy :: Proxy api) pc nt . s
+
   route Proxy context subserver = route (Proxy :: Proxy api) context $
       subserver `addParameterCheck` withRequest paramsCheck
     where
@@ -410,6 +522,8 @@ instance (KnownSymbol sym, HasServer api context)
   type ServerT (QueryFlag sym :> api) m =
     Bool -> ServerT api m
 
+  hoistServerWithContext _ pc nt s = hoistServerWithContext (Proxy :: Proxy api) pc nt . s
+
   route Proxy context subserver =
     let querytext r = parseQueryText $ rawQueryString r
         param r = case lookup paramname (querytext r) of
@@ -433,6 +547,8 @@ instance HasServer Raw context where
 
   type ServerT Raw m = Tagged m Application
 
+  hoistServerWithContext _ _ _ = retag
+
   route Proxy _ rawApplication = RawRouter $ \ env request respond -> runResourceT $ do
     -- note: a Raw application doesn't register any cleanup
     -- but for the sake of consistency, we nonetheless run
@@ -466,11 +582,13 @@ instance HasServer Raw context where
 -- > server = postBook
 -- >   where postBook :: Book -> Handler Book
 -- >         postBook book = ...insert into your db...
-instance ( AllCTUnrender list a, HasServer api context
-         ) => HasServer (ReqBody list a :> api) context where
+instance ( AllCTUnrender list a, HasServer api context, SBoolI (FoldLenient mods)
+         ) => HasServer (ReqBody' mods list a :> api) context where
 
-  type ServerT (ReqBody list a :> api) m =
-    a -> ServerT api m
+  type ServerT (ReqBody' mods list a :> api) m =
+    If (FoldLenient mods) (Either String a) a -> ServerT api m
+
+  hoistServerWithContext _ pc nt s = hoistServerWithContext (Proxy :: Proxy api) pc nt . s
 
   route Proxy context subserver
       = route (Proxy :: Proxy api) context $
@@ -491,9 +609,11 @@ instance ( AllCTUnrender list a, HasServer api context
       -- Body check, we get a body parsing functions as the first argument.
       bodyCheck f = withRequest $ \ request -> do
         mrqbody <- f <$> liftIO (lazyRequestBody request)
-        case mrqbody of
-          Left e  -> delayedFailFatal err400 { errBody = cs e }
-          Right v -> return v
+        case sbool :: SBool (FoldLenient mods) of
+          STrue -> return mrqbody
+          SFalse -> case mrqbody of
+            Left e  -> delayedFailFatal err400 { errBody = cs e }
+            Right v -> return v
 
 -- | Make sure the incoming request starts with @"/path"@, strip it and
 -- pass the rest of the request path to @api@.
@@ -506,32 +626,51 @@ instance (KnownSymbol path, HasServer api context) => HasServer (path :> api) co
       (cs (symbolVal proxyPath))
       (route (Proxy :: Proxy api) context subserver)
     where proxyPath = Proxy :: Proxy path
+  hoistServerWithContext _ pc nt s = hoistServerWithContext (Proxy :: Proxy api) pc nt s
 
 instance HasServer api context => HasServer (RemoteHost :> api) context where
   type ServerT (RemoteHost :> api) m = SockAddr -> ServerT api m
 
   route Proxy context subserver =
     route (Proxy :: Proxy api) context (passToServer subserver remoteHost)
+  hoistServerWithContext _ pc nt s = hoistServerWithContext (Proxy :: Proxy api) pc nt . s
 
 instance HasServer api context => HasServer (IsSecure :> api) context where
   type ServerT (IsSecure :> api) m = IsSecure -> ServerT api m
 
   route Proxy context subserver =
     route (Proxy :: Proxy api) context (passToServer subserver secure)
-
     where secure req = if isSecure req then Secure else NotSecure
 
+  hoistServerWithContext _ pc nt s = hoistServerWithContext (Proxy :: Proxy api) pc nt . s
+
 instance HasServer api context => HasServer (Vault :> api) context where
   type ServerT (Vault :> api) m = Vault -> ServerT api m
 
   route Proxy context subserver =
     route (Proxy :: Proxy api) context (passToServer subserver vault)
+  hoistServerWithContext _ pc nt s = hoistServerWithContext (Proxy :: Proxy api) pc nt . s
 
 instance HasServer api context => HasServer (HttpVersion :> api) context where
   type ServerT (HttpVersion :> api) m = HttpVersion -> ServerT api m
 
   route Proxy context subserver =
     route (Proxy :: Proxy api) context (passToServer subserver httpVersion)
+  hoistServerWithContext _ pc nt s = hoistServerWithContext (Proxy :: Proxy api) pc nt . s
+
+-- | Ignore @'Summary'@ in server handlers.
+instance HasServer api ctx => HasServer (Summary desc :> api) ctx where
+  type ServerT (Summary desc :> api) m = ServerT api m
+
+  route _ = route (Proxy :: Proxy api)
+  hoistServerWithContext _ pc nt s = hoistServerWithContext (Proxy :: Proxy api) pc nt s
+
+-- | Ignore @'Description'@ in server handlers.
+instance HasServer api ctx => HasServer (Description desc :> api) ctx where
+  type ServerT (Description desc :> api) m = ServerT api m
+
+  route _ = route (Proxy :: Proxy api)
+  hoistServerWithContext _ pc nt s = hoistServerWithContext (Proxy :: Proxy api) pc nt s
 
 -- | Singleton type representing a server that serves an empty API.
 data EmptyServer = EmptyServer deriving (Typeable, Eq, Show, Bounded, Enum)
@@ -551,6 +690,8 @@ instance HasServer EmptyAPI context where
 
   route Proxy _ _ = StaticRouter mempty mempty
 
+  hoistServerWithContext _ _ _ = retag
+
 -- | Basic Authentication
 instance ( KnownSymbol realm
          , HasServer api context
@@ -567,6 +708,8 @@ instance ( KnownSymbol realm
        basicAuthContext = getContextEntry context
        authCheck = withRequest $ \ req -> runBasicAuth req realm basicAuthContext
 
+  hoistServerWithContext _ pc nt s = hoistServerWithContext (Proxy :: Proxy api) pc nt . s
+
 -- * helpers
 
 ct_wildcard :: B.ByteString
@@ -591,3 +734,77 @@ instance (HasContextEntry context (NamedContext name subContext), HasServer subA
 
       subContext :: Context subContext
       subContext = descendIntoNamedContext (Proxy :: Proxy name) context
+
+  hoistServerWithContext _ _ nt s = hoistServerWithContext (Proxy :: Proxy subApi) (Proxy :: Proxy subContext) nt s
+
+-------------------------------------------------------------------------------
+-- TypeError helpers
+-------------------------------------------------------------------------------
+
+#ifdef HAS_TYPE_ERROR
+-- | This instance catches mistakes when there are non-saturated
+-- type applications on LHS of ':>'.
+--
+-- >>> serve (Proxy :: Proxy (Capture "foo" :> Get '[JSON] Int)) (error "...")
+-- ...
+-- ...Expected something of kind Symbol or *, got: k -> l on the LHS of ':>'.
+-- ...Maybe you haven't applied enough arguments to
+-- ...Capture' '[] "foo"
+-- ...
+--
+-- >>> undefined :: Server (Capture "foo" :> Get '[JSON] Int)
+-- ...
+-- ...Expected something of kind Symbol or *, got: k -> l on the LHS of ':>'.
+-- ...Maybe you haven't applied enough arguments to
+-- ...Capture' '[] "foo"
+-- ...
+-- 
+instance TypeError (HasServerArrowKindError arr) => HasServer ((arr :: k -> l) :> api) context
+  where
+    type ServerT (arr :> api) m = TypeError (HasServerArrowKindError arr)
+    -- it doens't really matter what sub route we peak
+    route _ _ _ = error "servant-server panic: impossible happened in HasServer (arr :> api)"
+    hoistServerWithContext _ _ _ = id
+
+-- Cannot have TypeError here, otherwise use of this symbol will error :)
+type HasServerArrowKindError arr =
+    'Text "Expected something of kind Symbol or *, got: k -> l on the LHS of ':>'."
+    ':$$: 'Text "Maybe you haven't applied enough arguments to"
+    ':$$: 'ShowType arr
+
+-- | This instance prevents from accidentally using '->' instead of ':>'
+--
+-- >>> serve (Proxy :: Proxy (Capture "foo" Int -> Get '[JSON] Int)) (error "...")
+-- ...
+-- ...No instance HasServer (a -> b).
+-- ...Maybe you have used '->' instead of ':>' between
+-- ...Capture' '[] "foo" Int
+-- ...and
+-- ...Verb 'GET 200 '[JSON] Int
+-- ...
+--
+-- >>> undefined :: Server (Capture "foo" Int -> Get '[JSON] Int)
+-- ...
+-- ...No instance HasServer (a -> b).
+-- ...Maybe you have used '->' instead of ':>' between
+-- ...Capture' '[] "foo" Int
+-- ...and
+-- ...Verb 'GET 200 '[JSON] Int
+-- ...
+--
+instance TypeError (HasServerArrowTypeError a b) => HasServer (a -> b) context
+  where
+    type ServerT (a -> b) m = TypeError (HasServerArrowTypeError a b)
+    route _ _ _ = error "servant-server panic: impossible happened in HasServer (a -> b)"
+    hoistServerWithContext _ _ _ = id
+
+type HasServerArrowTypeError a b =
+    'Text "No instance HasServer (a -> b)."
+    ':$$: 'Text "Maybe you have used '->' instead of ':>' between "
+    ':$$: 'ShowType a
+    ':$$: 'Text "and"
+    ':$$: 'ShowType b
+#endif
+
+-- $setup
+-- >>> import Servant
